/**
 * Copyright Dave Cline
 * Do not distribute
 */

class NoteController
{
  public:
    const static int MAX_VELOCITY = 127; //MIDI only accepts up to 127
    const static int MIN_VELOCITY = 1;   //No use sending a 0
    const static int SUSTAIN_NOTE_MS = 10;
    const static unsigned int TRAINING_MIN_MS_BETWEEN_NOTES = 1000;
    const static int USE_ROUNDUP_DIVISIONS = 1;
    const static int NUM_ROUNDUP_DIVISIONS = 12;
    const static int VELOCITY_BUCKET_SIZE = MAX_VELOCITY / NUM_ROUNDUP_DIVISIONS;
    const static int PERCENT_VELOCITY_MULTIPLICATION = 230.0f; //Set to 100.0f to use the calculated velocity as-is.  Use higher values to make everything louder (up to the max) and lower to make everything quieter (down to the min)
    const unsigned int TIME_BETWEEN_NOTES_MS = 100; //Minimum time to wait before sending notes in milliseconds.  Don't forget we also spend a certain amount of time waiting for a peizo to hit its peak reading (see the piezo code for how long)
    static const bool PLAY_HIGH_HAT_CLOSING_NOTE = true; //Set to true to make a tapping sound when the high hat is closed
    const int HIGH_HAT_OPEN_NOTE = 46;  
    const int HIGH_HAT_CLOSED_NOTE = 42;
    const int KICK_DRUM_NOTE = 36;

    //Midi notes beow correspond to the expected values for power drummer2 and dtxmania
    //Currently these are yellow (snare),  high hat (0), green(high tom),red (LOW TOM),  purple (crash cymbal)  
    const int MIDI_NOTE_NUMBERS[9] = { 38,  0, 48, 47, 49, 0, 0, 0 }; 

    int MIDI_CHAN = 1;
    long noteOnAt = 0;
    bool noteOffSent = false;
    int midiNote = 0;
    int noteIndex = 0;
    long numNotes = 0;
    float velocity = 0;
    bool isHighHat = false;
    bool isKickDrum = false;
    unsigned long millisSinceLastNoteOn = 0;

    NoteController(int);
    void tick();
    void noteOn(float,float,float,bool);
    void noteOff();
    void playHighHatClosingNote();
};

NoteController::NoteController(int noteIndex2)
{
  noteIndex = noteIndex2;
  midiNote = MIDI_NOTE_NUMBERS[noteIndex];
  MIDI_CHAN = noteIndex+1;
}

void NoteController::noteOn(float piezoVelocity, float minPiezoVelocity, float maxPiezoVelocity, bool isHighHatClosed)
{
  //Don't turn "on" more than one note at a time
  //This is more to prevent double hits than double-sounds.  You can lower time between notes
  //to a smaller value if you are having trouble getting the drumset to pick up your strikes,
  //but it may come at the cost of getting more double hits, depending on drumset setup
  if(noteOnAt != 0)
  {
    if(DEBUG_NOTE_CONTROLLER)
    {
      Serial.print("Can't turn on more than one note at a time.  Exiting noteOn for controller ");
      Serial.println(noteIndex);
    }

    return;
  }
  numNotes++;

  if(DEBUG_NOTE_CONTROLLER)
  {
    Serial.print(millis());
    Serial.print(" - Controller ");
    Serial.print(noteIndex);
    Serial.print("Num notes is now ");
    Serial.println(numNotes);
  }

  //Consider hitting the piezo harder or softer.  Each hit lands on a sliding scale.  Let's take
  //the percentage of total velocity that we know the piezo can reach and multiply it by the max velocity
  //that midi notes can reach.  In other words this is % of max of the piezo values X the max midi value
  velocity = MAX_VELOCITY * (piezoVelocity - minPiezoVelocity) / (maxPiezoVelocity - minPiezoVelocity); 

  //Sometimes we just want it louder or softer in general
  velocity = velocity * PERCENT_VELOCITY_MULTIPLICATION / 100.0f;

  //Round velocity up to a certain level.  How high the level is depends on the number of times we are
  //going to divide the maximum possible velocity.  3 divisions means this value will be 1/3,
  //2/3, or 3/3 max velocity
  if(USE_ROUNDUP_DIVISIONS)
  {
    div_t divOperationValue = div(velocity, VELOCITY_BUCKET_SIZE);
    float roundedUpVelocity = VELOCITY_BUCKET_SIZE * divOperationValue.quot;
    if(divOperationValue.rem > 0)
    {
      //roundedUpVelocity += VELOCITY_BUCKET_SIZE;
    }
    velocity = roundedUpVelocity;
  }

  //Don't bother sending velocities that are higher than the possible max as per MIDI standards
  if(velocity > MAX_VELOCITY)
  {
    if(DEBUG_NOTE_CONTROLLER)
    {
      Serial.print("Rounding down velocity ");
      Serial.println(velocity);
    }

    velocity = MAX_VELOCITY;
  }
  //Don't send velocities below the minimum
  else if(velocity < MIN_VELOCITY)
  {
    if(DEBUG_NOTE_CONTROLLER)
    {
      Serial.print("Rounding up velocity ");
      Serial.println(velocity);
    }

    velocity = MIN_VELOCITY;
  }

  //Always use max velocity... I can't get consistent velocity sensitivity on my piezos
  velocity = MAX_VELOCITY;

  //Select the correct midi node
  midiNote = MIDI_NOTE_NUMBERS[noteIndex];

  //If this is the high-hat, determine the midi note a little differently
  if(isHighHat)
  {
    if(isHighHatClosed)
    {
      midiNote = HIGH_HAT_CLOSED_NOTE;
    } else {
      midiNote = HIGH_HAT_OPEN_NOTE;
    }

    if(DEBUG_NOTE_CONTROLLER)
    {
      Serial.print("Overriding whatever note with high hat note ");
      Serial.println(midiNote);
    }  
  }

  if(isKickDrum)
  {
    midiNote = KICK_DRUM_NOTE;
    
    if(DEBUG_NOTE_CONTROLLER)
    {
      Serial.print("Overriding whatever note with kick drum note ");
      Serial.println(midiNote);
    }  
  }

  if(DEBUG_NOTE_CONTROLLER)
  {
    Serial.print(millis());
    Serial.print(" - controller ");
    Serial.print(noteIndex);
    Serial.print(" activating note ");
    Serial.print(midiNote);
    Serial.print(" with velocity ");
    Serial.println(velocity);
  }

  usbMIDI.sendNoteOn(midiNote, velocity, MIDI_CHAN);
  noteOnAt = millis();
  noteOffSent = false;
}

void NoteController::playHighHatClosingNote() 
{
  if(DEBUG_NOTE_CONTROLLER)
  {
    Serial.print(millis());
    Serial.println(" sending high hat closed sound on and probably also off");
  }
  usbMIDI.sendNoteOn(HIGH_HAT_CLOSED_NOTE, MIN_VELOCITY, MIDI_CHAN);

  if(SEND_NOTE_OFFS) 
  {
    usbMIDI.sendNoteOff(HIGH_HAT_CLOSED_NOTE, 0, MIDI_CHAN);
  }
}

void NoteController::tick()
{
  if(noteOnAt != 0)
  {
    if(!noteOffSent && millis() - noteOnAt > SUSTAIN_NOTE_MS)
    {
      if(SEND_NOTE_OFFS)
      {
        if(DEBUG_NOTE_CONTROLLER)
        {
          Serial.print(millis());
          Serial.print(" - controller ");
          Serial.print(noteIndex);
          Serial.println(" sending note off");
        }
  
        usbMIDI.sendNoteOff(midiNote, 0, MIDI_CHAN);
      }
      noteOffSent = true;
    }

    if(millis() - noteOnAt >TIME_BETWEEN_NOTES_MS)
    {
      if(DEBUG_NOTE_CONTROLLER)
      {
        Serial.print(millis());
        Serial.print(" - controller ");
        Serial.print(noteIndex);
        Serial.println(" ready for another note");
      }

      noteOnAt = 0;
    }
  }
}
