/**
 * Copyright Dave Cline
 * Do not distribute
 */

class Piezo
{
  public:

    const static int MIN_MICROS_TO_WAIT_FOR_PEAK = 5000;    //The minimum number of microseconds to read the analog input before triggering a midi-on event
    const static int MAX_MICROS_TO_WAIT_FOR_PEAK = 10000;    //The maximum number of microseconds to read the analog input before triggering a midi-on event
    const static int PEAK_BOTTOM_THRESHOLD_PERCENTAGE = 30; //The percent above the noise level that the peaking curve should drop below before we trigger an event

    const static int STATE_CHILLIN = 0;           //No really doing anything special
    const static int STATE_READING_MAX = 1;       //The state of reading inputs and keeping track of the max, as we are about to send a midi signal with whatever the max turns out to be
    const static int STATE_READING_MIN = 2;       //The state of reading inputs after a peak, until we get back to a certain threshold between the peak and base noise reading
    
    
    int state = STATE_CHILLIN;
    unsigned long microsAtReadStart = 0;
    int numRisingSamples = 0;
    int numPeakSamples = 0;
    float peakSamplesTotal = 0;
    float minReading = 100;       //The minimum reading we saw during setup time when the piezo was at rest
    float maxReading = 0;         //The maximum reading we've seen so far. 
    float analogValue = 0;
    float recentValley = 9999;
    float recentPeak = 0;
    
    Piezo(int);
    float tick();
    
  private:
    int analogPinIndex = 0;
    float read();
};

Piezo::Piezo(int analogPinIndex2)
{
  analogPinIndex = analogPinIndex2;
}


float Piezo::tick()
{
  analogValue = read();

  //If we're not doing anything at the moment and it seems like an input peak is appearing
  if(state == STATE_CHILLIN)
  {
    if(analogValue > minReading + NOISE_BUFFER)
    {
      //Start reading to see when we hit a local maximum
      state = STATE_READING_MAX;

      //Reset vars 
      microsAtReadStart = micros();
      numRisingSamples = 1;
      numPeakSamples = 0;
      recentPeak = analogValue;
      recentValley = analogValue;
      peakSamplesTotal = 0;
    }
    //Get out of here as quickly as possible if we have no reason to do anything 
    //(which is the case if we don't have a pulse to read)
    else
    {
      return 0;
    }
  }

  //If we're currently polling the input on the pin to see how high it goes...
  if(state == STATE_READING_MAX)
  {
    numPeakSamples++;

    peakSamplesTotal += analogValue;

    //Keep track of highest point of recent peak(s)
    if(analogValue > recentPeak)
    {
      numRisingSamples++;
      recentPeak = analogValue;  

      if(DEBUG_PIEZO_PEAK)
      {
        Serial.print("Peak found rising.  Peak max so far is: ");
        Serial.println(recentPeak);
      }
    }
    //Keep track of the lowest point
    if(analogValue < recentValley)
    {
      recentValley = analogValue;

      if(DEBUG_PIEZO_PEAK)
      {
        Serial.print("Valley found falling.  Valley value is: ");
        Serial.println(recentValley);
      }
    }

    //This is as peaked as we're gonna get.  Let's stop listening and return what we've found so far
    if((recentPeak > minReading + NOISE_BUFFER //(analogValue < minReading + NOISE_BUFFER + PEAK_BOTTOM_THRESHOLD_PERCENTAGE * (recentPeak - minReading - NOISE_BUFFER) / 100.0f
    && micros() - microsAtReadStart > MIN_MICROS_TO_WAIT_FOR_PEAK)
    || micros() - microsAtReadStart > MAX_MICROS_TO_WAIT_FOR_PEAK)
    {
      state = STATE_CHILLIN; //Go back to being ready for a new peak
      float averagePeakValue = peakSamplesTotal / numPeakSamples;

      //Return the peak of the rising samples that we found over the last however many millis
      if(DEBUG_PIEZO_PEAK)
      {
        Serial.print("Top of actual peak found.  Average was: ");
        Serial.print(averagePeakValue);
        Serial.print(" Maximum was :");
        Serial.println(recentPeak);
      }

      //Pause after hyper-reading peak so that we can see the wave form
      if(DEBUG_PIEZO_PEAK_SLOWMO)
      {
        Serial.println(minReading);
        Serial.println(minReading);
        Serial.println(minReading);
        Serial.println(minReading);
        Serial.println(minReading);
        Serial.println(minReading);
        Serial.println(recentPeak);
        Serial.println(recentPeak);
        Serial.println(recentPeak);
        Serial.println(recentPeak);
        Serial.println(recentPeak);
        Serial.println(recentPeak);
        Serial.println(recentPeak);
        Serial.println(maxReading);
        Serial.println(maxReading);
        Serial.println(maxReading);
        Serial.println(maxReading);
        //Serial.println(minReading + PEAK_BOTTOM_THRESHOLD_PERCENTAGE * (recentPeak - minReading) / 100.0f);

        delay(1000);          
      }

      return recentPeak;  
    }
    //Read the values from this piezo very rapidly
    else if(DEBUG_PIEZO_PEAK_SLOWMO)
    {
      delayMicroseconds(1500);
      Serial.println("");
      tick();
      
    }
  }

  return 0.0f;
}

float Piezo::read()
{
  //Read from the analog pin
  analogValue = analogRead(analogPinIndex);
   
  if(analogValue > maxReading)
  {
    maxReading = analogValue;

    if(DEBUG_PIEZO_MAX)
    {
      Serial.print("Maximum reading so far for piezo ");
      Serial.print(analogPinIndex);
      Serial.print(" is now ");
      Serial.println(maxReading);
    }
  }

  if(DEBUG_PIEZO) {
    Serial.print(analogValue);
    Serial.print("\t");
  }

  return analogValue;
}
