/**
 * Copyright Dave Cline
 * Do not distribute
 *
 * Note that this code is mostly set up for analog inputs, then had a couple of digital inputs sort of tacked on
 * So if you see "input" without "analog" or "digital" specified, then most likely it refers to an analog input
 */

//Debugging flags.  Note:  some of these are expected to work with the serial monitor, and some with the serial
//plotter.  You can't use both the plotter and the monitor at the same time though.  Use plotter to graph values (e.g. if debugging piezo waveforms), 
//and monitor to see text (e.g. if debugging when notes are sent from the arduino)
const int DEBUG_MAIN_SETUP = 0;
const int DEBUG_MAIN_LOOP = 0;
const int DEBUG_NOTE_CONTROLLER = 1;
const int DEBUG_DIGITAL_VALUES = 0;
const int DEBUG_PIEZO_MAX = 0;    //Set to 1 to print out messages when a new highest value is found for a specific piezo
const int DEBUG_PIEZO_PEAK = 0;   //Set to 1 to show wave form created by piezo when you load up the serial plotter

//Set to 1 to plot piezo values on graph
const int DEBUG_PIEZO = 0;        
const int DEBUG_PIEZO_PEAK_SLOWMO = 0; //Use when you need to figure out the wave form of your piezo output.  This will probably muck up MIDI, so keep this off normally

//General constants
const int NUM_INPUTS = 5;                    //The number of analog inputs.  Does not include digital inputs
const int MAX_NOISE = 100;                   //TURN THIS WAY UP IF YOU HAVEN'T SOLDERED YOUR CIRCUITRY. The amount of noise to tolerate in setup when looking for noise-only pins
const int MAX_NOISY_READS = 10;              //The number of noisy readings we can get from an unstruck input before ignoring it permanently (until restart of arduino)
const int NUM_NOISE_SAMPLES = 100;           //The number of initial readings to take while we determine if a piezo is just too noisy to pay attention to or not
const int DELAY_BETWEEN_NOISE_SAMPLES = 5;   //Milliseconds, I think.  This is for the setup function not the between-strike minimum delay
const int NOISE_BUFFER = 500;                //TURN THIS WAY UP IF YOU HAVEN'T SOLDERED YOUR CIRCUITRY.  Or if you plot your piezo values and turn on DEBUG_PIEZO, and see a lot of wavy lines going above the 50 mark, turn it up then too.  
                                             //This value + the maximum reading found at-rest for a piezo, is the minimum value we need to see on a piezo before we do something about it
const int SEND_NOTE_OFFS = 1;                //Generally with drums you don't actually need to send note-off midi signals, but you can if you wanna.  Setting this to 1 just causes them to be sent once the elapsed wait time between notes is over, with the exception of the high hat closing sound, which sends off immediately after on

//Include these after variables, because they may use them
#include "Piezo.h"
#include "HighHatController.h"
#include "NoteController.h"
#include "KickDrumController.h"


//Variables
int i,j,peak=0;
bool ignorePin[NUM_INPUTS], isHighHatClosed;
Piezo* piezos[NUM_INPUTS];
NoteController* noteControllers[NUM_INPUTS];
NoteController* kickDrumNoteController;
HighHatController* highHatController;
KickDrumController* kickDrumController;

void setup()
{
  Serial.begin(38400);
  analogReadResolution(12);

  //For the bass pedal digital inputs.  I use a 2-pin pushbutton connected to a digital pin on one end
  //and to digital ground on the other.  When the button is pressed, it allows the electrons to
  //drain to ground from the input pin to ground.  We achieve this by "writing" electrons
  //to the input pin, then look for when the signal drops to 0
  pinMode(KickDrumController::BASS_PEDAL_PIN, INPUT);
  digitalWrite(KickDrumController::BASS_PEDAL_PIN, HIGH);
  pinMode(HighHatController::HIGH_HAT_PEDAL_PIN, INPUT);
  digitalWrite(HighHatController::HIGH_HAT_PEDAL_PIN, HIGH);

  //Initialize variables
  int numNoisyReads[NUM_INPUTS];   //The number of times we read a pin and saw noise
  float maxNoisyRead[NUM_INPUTS];  //The highest noisy read we found
  for(i = 0; i < NUM_INPUTS; i++)
  {
    numNoisyReads[i] = 0;
    maxNoisyRead[i] = 0;
    ignorePin[i] = false;
  }

  //Listen for noisy pins, so we can ignore them in the main loop
  float analogValue;
  for(i = 0; i < NUM_NOISE_SAMPLES; i++)
  {
    for(j = 0; j < NUM_INPUTS; j++)
    {
      analogValue = analogRead(j);

      //Output the noise
      if(DEBUG_MAIN_SETUP && analogValue > MAX_NOISE)
      {
        Serial.print("Pin ");
        Serial.print(j);
        Serial.print(" read ");
        Serial.println(analogValue);
      }
      //Keep track of the highest analog value we've seen so far
      if(analogValue > maxNoisyRead[j])
      {
        if(DEBUG_MAIN_SETUP)
        {
          Serial.print("Max noise reading for ");
          Serial.print(j);
          Serial.print(" is now ");
          Serial.println(analogValue);
        }
        maxNoisyRead[j] = analogValue;
      }
      if(analogValue > MAX_NOISE)
      {
        numNoisyReads[j]++;
      }
    }

    //Spread out our reads so we don't get a bunch of samples during a brief period of
    //time that just happened to be quieter than usual
    delay(DELAY_BETWEEN_NOISE_SAMPLES);
  }

  //Figure out which pins are too noisy to pay attention to
  //and set a flag that says to ignore them
  for(i = 0; i < NUM_INPUTS; i++)
  {
    if(numNoisyReads[i] > MAX_NOISY_READS)
    {
      ignorePin[i] = true;

      if(DEBUG_MAIN_SETUP)
      {
        Serial.print("Ignoring noisy pin ");
        Serial.print(i);
        Serial.print(" due to ");
        Serial.print(numNoisyReads[i]);
        Serial.print(" noisy reads, with max ");
        Serial.println(maxNoisyRead[i]);
      }
    }
  }

  //Create objects for all analog pins that we're not ignoring due to noise (excessive noise usually means there aren't any wires plugged in
  //at those pins, but even if there are wires, excessively noisy inputs are going to fase-trigger all the time and ruin everything anyways,
  //might as well ignore them regardless)
  for(i = 0; i < NUM_INPUTS; i++)
  {
    if(!ignorePin[i])
    {
      noteControllers[i] = new NoteController(i);
      piezos[i] = new Piezo(i);
      piezos[i]->minReading = maxNoisyRead[i];
      piezos[i]->maxReading = piezos[i]->minReading;

      if(DEBUG_MAIN_SETUP)
      {
        Serial.print("Listening on pin ");
        Serial.print(i);
        Serial.print(" with an acceptable ");
        Serial.print(numNoisyReads[i]);
        Serial.print(" noisy reads.  The minimum reading to register a hit will be ");
        Serial.println(piezos[i]->minReading + NOISE_BUFFER);
      }
    }
  }
  highHatController = new HighHatController();
  noteControllers[HighHatController::HIGH_HAT_ANALOG_PIN]->isHighHat = true;
  kickDrumNoteController = new NoteController(0);
  kickDrumNoteController->isKickDrum = true;

  if(DEBUG_MAIN_SETUP)
  {
    Serial.println("Done in setup");
  }
}

void loop()
{
  //Process high hat foot pedal
  int justClosed = highHatController->tick();
  if(justClosed) 
  {
    if(DEBUG_MAIN_LOOP)
    {
      Serial.println("Just closed high hat ");
    }
    
    //Send "high hat pedal closed high hat" note, usually the same as a tap sound but you can make it whatever
    if(NoteController::PLAY_HIGH_HAT_CLOSING_NOTE)
    {
      if(DEBUG_MAIN_LOOP)
      {
        Serial.println("Playing closed high hat note sound");
      }

      noteControllers[HighHatController::HIGH_HAT_ANALOG_PIN]->playHighHatClosingNote();
    }
  }

  //Process kick drum foot pedal
  int bassValue = kickDrumController->tick();
  if(bassValue == KickDrumController::BASS_KICKED_VALUE && kickDrumNoteController->noteOnAt == 0)
  {
    if(DEBUG_MAIN_LOOP)
    {
      Serial.print("kick drum input calling noteon with raw reading of ");
      Serial.print(bassValue);
      Serial.print(" and adjusted reading of ");
      Serial.println(1.0f);
    }

    kickDrumNoteController->noteOn(1.0f, 1.0f, 1.0f, 0); //Replace with real values if you ever have a velocity sensitive kick drum
  }
  kickDrumNoteController->tick();
  
  //Process analog inputs
  //Loop through all of our analog inputs
  for(i = 0; i < NUM_INPUTS; i++)
  {
    //Ignore pins that are reading random noise
    if(ignorePin[i])
    {
      continue;
    }
    
    //See what value we have on our piezo.  Trigger a note if the pulse is higher than the min reading and
    //we aren't currently already playing a note
    peak = piezos[i]->tick();
    if(peak >= piezos[i]->minReading + NOISE_BUFFER && noteControllers[i]->noteOnAt == 0)
    {
        if(DEBUG_MAIN_LOOP)
        {
          Serial.print("Piezo ");
          Serial.print(i);
          Serial.print(" calling noteon with raw reading of ");
          Serial.print(peak);
        }

        noteControllers[i]->noteOn(peak, piezos[i]->minReading, piezos[i]->maxReading, highHatController->isHighHatClosed);
    }

    //Allows the controller to turn off midi notes based on elapsed time, etc
    noteControllers[i]->tick();
  }

  //Need this here because debugging the peizo will output a bunch of tabs between piezo reads
  if(DEBUG_PIEZO)
  {
    Serial.println(0);
  }

  //Make it slightly easier to tell what our piezo waves look like
  if(DEBUG_PIEZO || DEBUG_PIEZO_PEAK_SLOWMO)
  {
    delay(1);
  }
}

